import logging
from scrapy.exceptions import NotConfigured
from scrapy import signals
logger = logging.getLogger(__name__)

class CustomLogExtension:

    def __init__(self):
        self.level = logging.WARNING
        self.modules = ['scrapy.utils.log', 'scrapy.middleware',
                        'scrapy.extensions.logstats', 'scrapy.statscollectors', 
                        'scrapy.core.engine', 'scrapy.core.scraper', 
                        'scrapy.crawler', 'scrapy.extensions', 
                        __name__]
        for module in self.modules:
            logger = logging.getLogger(module)
            logger.setLevel(self.level)

    @classmethod
    def from_crawler(cls, crawler):
        if not crawler.settings.getbool('CUSTOM_LOG_EXTENSION'):
            raise NotConfigured
        ext = cls()
        crawler.signals.connect(
            ext.spider_opened, signal=signals.spider_opened
        )
        return ext

    def spider_opened(self, spider):
        logger.debug("This log should not appear.")


# useful for handling different item types with a single interface
from itemadapter import ItemAdapter
from datetime import datetime
import re
from scrapy import signals
from scrapy.signalmanager import dispatcher


def clean_number(number):

    if number is None:
        return ""
    
    cleaned_number = number.replace("€","").replace(",","").replace('','').strip()

    return cleaned_number



class MergedDataPipeline:
    def __init__(self):
        self.raw_data = []
        self.cleaned_data = []
        self.list_dic = {}

    def process_item(self, item, spider):
        adapter = ItemAdapter(item)

        if adapter.get('source'):
            
            #Process raw data
            self.raw_data.append(adapter.asdict())

            #Process cleaned data
            cleaned_item = self.clean_item(item)
            self.cleaned_data.append(cleaned_item)

        return item
    
    def close_spider(self, spider):
        # Convert values to list for keys in list_dic
        for key in self.list_dic:
            for cleaned_item in self.cleaned_data:
                self.convert_to_list(cleaned_item, key)

        # Process data
        spider.log("Closing spider and sending custom_closed_signal...")
        dispatcher.send(
            signal=signals.spider_closed,
            sender=spider,
            cleaned_data=self.cleaned_data,
            raw_data=self.raw_data,
            list_dic=self.list_dic
        )

    
    def convert_to_list(self, cleaned_item, key):

        # Recursive function to convert values to list for a specific key
        if isinstance(cleaned_item, dict):
            for sub_key, sub_value in cleaned_item.items():
                if sub_key == key and sub_key in self.list_dic:
                    cleaned_item[sub_key] = [sub_value] if not isinstance(sub_value, list) else sub_value
                elif isinstance(sub_value,(dict, list)):
                    self.convert_to_list(sub_value, key)
        elif isinstance(cleaned_item, list):
            for element in cleaned_item:
                self.convert_to_list(element, key)

    def clean_item (self,item):

        if isinstance(item, dict):
            cleaned_item = {}
            for key, value in item.items():
                cleaned_value = self.clean_value(key, value)
                if cleaned_value is not None:
                    cleaned_item[key] = cleaned_value
            return cleaned_item
        elif isinstance(item,list):
            cleaned_list = []
            for element in item:
                cleaned_element = self.clean_item(element)
                if cleaned_element:
                    cleaned_list.append(cleaned_element)
            return cleaned_list
        else:
            return item
    
    def clean_value(self, key, value):

        #the keys that need to be in float format
        float_keys = ['residence_charge','water','deposit','rent','living_area','sauna','deposit','right_of_occupancy_fee','other_area','total_area','other_heating','electric_heating','electricity','parking_space','tv_cable_tv']
        #the keys that need to be in int format
        int_keys = ['number_of_rooms','number_of_roommates','rental_number','renovation_year','construction_year','toiler_count','phone_connections','tv_connections','floor']
        #the keys that need to be in timestamp
        date_keys = ['inspection_done','vacancy']
        #the keys that are special and not devided
        special_str_keys = ['description_of_rental','more_information','special_terms','information_about_costs','further_information','additional_information_about_energy_certificate','name','maintenance','material_description','yard_description','room_description','other_info','past_renovation','driving_instructions','renovations','transportation','description']

        if isinstance(value, str):
            #float
            if key in float_keys:
                value = clean_number(value)

                match = re.search(r'(\d[\d,]*)', value)
                numeric_value = float(match.group().replace(',', '')) if match else None

                # Extract frequency information
                if '/w' in value or '/week' in value:
                    frequency = 'per_week'
                elif '/month'in value or '/m' in value:
                    frequency = 'per_month'
                elif '/year' in value or '/y' in value:
                    frequency = 'per_year'
                else:
                    frequency = None

                # Extract per_person information
                per_person = 'per_person' if '/person' in value else None

                # Update the cleaned_data dictionary
                cleaned_value = {
                    f"{key}_value": numeric_value,
                    f"{key}_frequency": frequency,
                    f"{key}_person": per_person
                }

                return cleaned_value
                        
            #int
            elif key in int_keys:
                try:
                    return int(re.search('\d+',value).group())
                except(ValueError,AttributeError):
                    return value
            #date timestamp
            elif key in date_keys:
                #convert to timestamp
                if value == "vacant immediately" or value == "on agreement":
                    return value
                else:
                    try:
                        date_formts = ["%m/%d/%y"]
                        for date_format in date_formts:
                            try:
                                value = datetime.strptime(value,date_format)
                                unix_timestamp = int(value.timestamp())
                                return unix_timestamp
                            except ValueError:
                                pass
                    except Exception as e:
                        return None
                    
            # convert into list
            elif key not in special_str_keys:

                try:
                    parts = value.split(',')
                    if len(parts) > 1:
                        self.list_dic[key] = None
                        return [part.strip() for part in parts]
                    else:
                        return value
                except AttributeError:
                    return value
            else:
                return value
            
        elif isinstance(value, dict):
            # Handle nested dictionaries (e.g., house_type or house data)
            cleaned_dict = {}
            for sub_key, sub_value in value.items():
                cleaned_sub_value = self.clean_value(sub_key, sub_value)
                if cleaned_sub_value is not None:
                    cleaned_dict[sub_key] = cleaned_sub_value
            return cleaned_dict
    
        elif isinstance(value, list):
            # Handle nested lists (e.g., list of house data)
            cleaned_list = []
            for element in value:
                cleaned_element = self.clean_value(None, element)
                if cleaned_element:
                    cleaned_list.append(cleaned_element)
            return cleaned_list
        
        else:
            return value
        


settings = {
   "BOT_NAME": "huurdatascraper",
   "SPIDER_MODULES": ["ClappScrapers.den_rental.spider"],
   "NEWSPIDER_MODULE": "ClappScrapers.den_rental.spider",
   "ROBOTSTXT_OBEY": True,
   "ITEM_PIPELINES": {
       "ClappScrapers.den_rental.spider.pipelines.MergedDataPipeline": 100,
   },
   "DOWNLOADER_MIDDLEWARES": {
       "scrapy.downloadermiddlewares.useragent.UserAgentMiddleware": None,
       "scrapy_user_agents.middlewares.RandomUserAgentMiddleware": 700,
   },
   "AUTOTHROTTLE_ENABLED": True,
   "AUTOTHROTTLE_START_DELAY": 0.5,
   "AUTOTHROTTLE_TARGET_CONCURRENCY": 2.0,
   "REQUEST_FINGERPRINTER_IMPLEMENTATION": "2.7",
   "TWISTED_REACTOR": "twisted.internet.asyncioreactor.AsyncioSelectorReactor",
   "FEED_EXPORT_ENCODING": "utf-8",
    "CUSTOM_LOG_EXTENSION":True,
    "EXTENSIONS":{
        'scrapy.extensions.telnet.TelnetConsole': None,
        'ClappScrapers.den_rental.spider.extension.CustomLogExtension': 1,}
}


import scrapy
from scrapy.exceptions import CloseSpider
import re
import time
from urllib.parse import urljoin



def clean_text(text):

    if text is None:

        return '-'

    if text is not None:

        cleaned_value = re.sub(r'<[^>]*>', '', text)
                
        # Remove leading and trailing whitespaces, including newline characters
        cleaned_value = cleaned_value.strip()

        # Replace consecutive newline characters with a single space
        cleaned_value = re.sub(r'\n+', ' ', cleaned_value)

        # Replace other unwanted characters
        cleaned_value = cleaned_value.replace("\xa0", "").replace("  ","").replace("m<sup>2</sup>","").replace("m<sup>3</sup>","").replace("\r","").replace("\t","").strip()


        # Use strip() to remove leading and trailing whitespace and newline characters
        return cleaned_value

def normalize_key (text):

    if text is None:
        return ''
    
    if text is not None:
        

        normalized_key = text.replace('(p.m.)','').replace('(s)','').strip()

        # Replace spaces with underscores
        normalized_key = normalized_key.replace(' ', '_').replace('.','').strip()
    
        # Remove special characters
        normalized_key = ''.join(char for char in normalized_key if char.isalnum() or char in ['_', '-'])

        normalized_key = normalized_key.replace('__',"_")

        # Decapitalize the key
        normalized_key = normalized_key.lower()

        return normalized_key



class finlandrentalspiderSpider(scrapy.Spider):
    
    name = "finlandrentalspider"
    start_urls = ["https://www.vuokraovi.com/vuokra-asunnot?locale=en&page=1&pagetype="]
    base_url = "https://www.vuokraovi.com/vuokra-asunnot?locale=en&page={}&pageType="

    custom_settings = {
        'RETRY_HTTP_CODES': [500, 502, 503, 504, 400, 403, 404, 408],
        'RETRY_TIMES': 5,
        'RETRY_PRIORITY_ADJUST': -1,        
        'DOWNLOAD_DELAY' : 2,  # 2 second delay between requests
        'HTTPERROR_ALLOW_ALL' : True,
    }


    all_possible_property_keys = set()

    max_retries = 3  # Set the maximum number of retries

    retry_counter = 0


    def parse(self, response):

                    
        property_list = []

        #initialize a set for all the possible keys in project_info to keep the amount of the keys for each dict project_info the same

        #Find all the properties
        properties = response.css('div.list-item-container div.row.top-row a.list-item-link[onclick^="setScrollPositionInCookie()"]')

        last_button = response.css('a.list-pager-button').getall()[-1]
        alt_content = scrapy.Selector(text=last_button).css('::attr(alt)').get()

        page_match = re.search(r'page=(\d+)', response.url)

        current_page = int(page_match.group(1))

        # Calculate the next page number with an increment of 1
        next_page_number = current_page + 1

        # Construct the next page URL
        next_page_link = self.base_url.format(next_page_number)
        
        if alt_content == 'Previous':

            raise CloseSpider("Reach the last page")

        if not properties:

            if self.retry_counter >= self.max_retries:

                self.logger.warning("Max retries reached, next page...")
                
                self.retry_counter = 0

                yield scrapy.Request(url=next_page_link, callback=self.parse)

            else:
                
                self.retry_counter += 1

                self.logger.warning("No properties found, waiting and retrying on the same page... (Retry {}/{}))".format(self.retry_counter, self.max_retries))

                time.sleep(0.5)

                # Make a request to the next page
                yield scrapy.Request(url=response.url, callback=self.parse)

        
        if properties:

            self.retry_counter = 0
        
        for property in properties :

            property_info= {}

            property_list.append(property_info)

            #Extract project links (for spiders to crawl into)

            property_link = property.css('a ::attr(href)').get()

            english_property_url = urljoin(property_link, '?locale=en')

                            
            if property_link:
                    
                yield scrapy.Request(url='https://www.vuokraovi.com' + english_property_url, callback=self.parse_property_page,cb_kwargs={'property_info':property_info})        
        
        else:

            yield scrapy.Request(url=next_page_link, callback=self.parse)
        

    
    def parse_property_page(self,response, property_info= None):



        try:


            property_info['source'] = response.url

            property_info['description_of_rental'] = clean_text(''.join(response.css('div#itempageDescription p::text').getall()))

            property_info['location'] = clean_text(' '.join(response.css('div.panel-body table tbody tr td a::text, div.panel-body table tbody tr td span::text').getall()[0:4]))


            property_columns = response.css('div.panel-body table tbody th::text').getall()[1:-1]
            property_columns = [normalize_key(i) for i in property_columns]

            self.all_possible_property_keys.update(property_columns)

            property_values = response.xpath('//div[@class="panel-body"]//table//tbody//td').getall()[1:-1]
            property_values = [clean_text(i) for i in property_values]
            property_values = [(i) for i in property_values if i != '' and i != ',']


            count = 0
            for column in property_columns:
                property_info[column] = property_values[count]
                count += 1

            property_info['information_url'] = response.css('td ul li a::attr(href)').getall()

            if property_info['description_of_rental'] == "":

                if not hasattr(self,'retry_count_property_page'):

                    self.retry_count_property_page = 0

                if self.retry_count_property_page < 2:

                    self.retry_count_property_page += 1
                    self.logger.warning("Page not loaded while parsing property page. Retrying the same page(Retry{}/2)...".format(self.retry_count_property_page))

                    yield scrapy.Request(url=response.url, callback=self.parse_property_page, cb_kwargs={'property_info':property_info})
                else:
                    self.retry_count_property_page = 0
                    self.logger.warning("Max retries reached for property page, moving on to the next property...")
                    return
            

            yield property_info

        except (IndexError, ValueError):

            # Catch IndexError and retry the same page
            if not hasattr(self, 'retry_counter_property_page'):
                self.retry_counter_property_page = 0

            if self.retry_counter_property_page < 2:
                self.retry_counter_property_page += 1
                self.logger.warning("IndexError while parsing property page. Retrying the same page (Retry {}/2)...".format(self.retry_counter_property_page))
                yield scrapy.Request(url=response.url, callback=self.parse_property_page, cb_kwargs={'property_info': property_info})
            else:
                self.retry_counter_property_page = 0
                self.logger.warning("Max retries reached for property page, moving on to the next property...")
                return
        




    def closed(self, reason):
        # This method is called when the spider is finished

        # Print all possible property keys
        self.log("All possible property keys:")
        for key in self.all_possible_property_keys:
            self.log(key)


import scrapy
from scrapy.http import TextResponse
import re
from scrapy.signalmanager import dispatcher
from scrapy import signals
import logging

logging.getLogger('scrapy').setLevel(logging.ERROR)


def clean_text(text):

    if text is None:

        return ''

    if text is not None:
                
        # Remove leading and trailing whitespaces, including newline characters
        cleaned_value = text.strip()

        # Replace consecutive newline characters with a single space
        cleaned_value = re.sub(r'\n+', ' ', cleaned_value)

        # Replace other unwanted characters
        cleaned_value = cleaned_value.replace("€", "").replace("\xa0", "").replace("  ","").replace("<td>","").replace("</td>","").replace('<td class="text-nowrap">',"").replace("m<sup>2</sup>","").replace("m<sup>3</sup>","").replace("-","").replace("<dd>","").replace("</dd>","").strip()


        # Use strip() to remove leading and trailing whitespace and newline characters
        return cleaned_value
    
def normalize_key (text):

    if text is None:
        return ''
    
    if text is not None:
        

        normalized_key = text.replace('(p.m.)','').replace('(s)','').strip()

        # Replace spaces with underscores
        normalized_key = normalized_key.replace(' ', '_').replace('.','').strip()
    
        # Remove special characters
        normalized_key = ''.join(char for char in normalized_key if char.isalnum() or char in ['_', '-'])

        # Decapitalize the key
        normalized_key = normalized_key.lower()

        return normalized_key


class NieuwbouwspiderSpider(scrapy.Spider):
    name = "Nieuwbouwspider"
    start_urls = ["https://www.nieuwbouw-in-amsterdam.nl/projecten/"]
    
    download_delay = 0.5  # 0.5 second delay between requests

    #set up a set for all the possible project keys
    all_possible_project_keys = set ()

    all_possible_house_type_keys = set()




    def parse(self, response):

            projects = response.css('a.card.card-project.card-project-list')


            #Extract project links (for spiders to crawl into)
        
            project_list = []

    
            for project in projects:

                project_info= {}

                project_link = project.css('a[href^="/project/"]::attr(href)').get()

                project_data = project.css('a.card.card-project.card-project-list').get()

                # Assuming project_data is a string containing  HTML

                project_data_bytes = project_data.encode('utf-8')

                project_data_html = TextResponse(url='dummyurl', body=project_data_bytes, encoding='utf-8')

                titles = project_data_html.css('svg[title]::attr(title)').getall()

                values = project_data_html.css('ul.list-inline li.list-inline-item span::text, ul.list-inline li.list-inline-item::text').getall()

                values = [clean_text(i) for i in values if i !=' ']

                titles = [normalize_key(i) for i in titles]

                count = 0
                for title in titles:
                    project_info[title] = values[count]
                    self.all_possible_project_keys.add(title)
                    count+=1

                if project_link:


                    yield response.follow(project_link, callback=self.parse_project_page,cb_kwargs={'projects':projects,'project':project,'project_info':project_info,'project_list':project_list})




    def parse_project_page(self, response, projects = None, project = None, project_list = list, project_info = None):

        # Extract information from the project


        #make the img_urls into a list instead of dict
        self.all_possible_project_keys.add("project_img_urls")
        project_imgs = response.css('img.img-fluid.w-100::attr(src)').getall()
        if project_imgs:
            project_info["project_img_urls"] = project_imgs


        labels = response.css('div.feature-label::text').getall()
        values = response.css('div.feature-value::text').getall()

        
        values = [clean_text(i) for i in values if i !=' ']
        labels = [normalize_key(i) for i in labels]

        if labels:
            count = 0
            for label in labels:
                 project_info[label] = values[count]
                 self.all_possible_project_keys.add(label)
                 count += 1
        
        self.all_possible_project_keys.add("over_dit_project") 

        project_description = response.css('section#project-beschrijving div.text-collapse p')
        if project_description :
             project_info ["over_dit_project"] = clean_text("".join(project_description.css('::text').getall()))


        self.all_possible_project_keys.add("project_source")   
        
        project_info['project_source'] = response.url

        project_info['stad'] = self.get_city_from_url(response.url)

        project_col = response.css('dl dt::text').getall()
        project_val = response.css('dl dd').getall()


        project_val = [clean_text(i) for i in project_val]

        #remove all the spaces inside the project_val 
        project_val = [i for i in project_val if i !='']

        project_col = [normalize_key(i) for i in project_col]

        #pair the project_col and project_val up in order
        count = 0
        for column_name in project_col:
            project_info[column_name] = project_val[count]
            self.all_possible_project_keys.add(column_name)
            count += 1

        #for project Type woningen make it into a list of strings in the pipeline 


        map_div = response.css('div#map')
        latitude = map_div.attrib['data-gemy']
        longitude = map_div.attrib['data-gemx']

        #separate the latitude and longitude
        project_info['locatie'] = { "latitude":latitude , "longitude":longitude }
        

        house_types = response.css('div.col-md-6.d-md-flex.d-lg-block')

        
        house_type_info_list = []

        project_info['house_types'] = house_type_info_list

        
        

        expected_house_types = len(house_types)

        found_house_types = False




        if house_types:
            

            house_type_links = house_types.css('a[href^="/type/"]::attr(href)').getall()
        

            for house_type_link in house_type_links:
                
                
                yield response.follow(house_type_link, callback=self.parse_house_type_page,cb_kwargs={'project_list':project_list,'project_info':project_info,'house_type_info_list':house_type_info_list,'expected_house_types':expected_house_types,'house_types':house_types})
                found_house_types = True

        #check if house_types were found
        if not found_house_types:
            yield project_info

            


#set up a check point for the situation if there is no sepicfic information for house type collections



    def parse_house_type_page(self,response, project = None, projects = None, project_list = list, project_info = None, house_type_info_list = list,expected_house_types = int, house_types = list):
        #Extract information from the collection

        house_type_data = response.css('table.table.table-responsive-md')

        #creat a list of house_info
        house_info_list = []
        #creat a lib called house_info
        house_info = {}

        self.all_possible_house_type_keys.update(['house_type_source','house_type_name'])
        house_type_info = {
                    'house_type_source':response.url,
                    'house_type_name':clean_text(response.css('h1.h2.card-title.font-weight-light.mb-4::text').get(default='')),
                }

        self.all_possible_house_type_keys.add("house_type_img_urls")
        house_type_imgs = response.css('img.img-fluid.w-100::attr(src)').getall()
        if house_type_imgs:
            house_type_info["house_type_img_urls"] = house_type_imgs

        self.all_possible_house_type_keys.add("over_dit_woningentype")
        house_type_description = response.css('div.col-md-10.text-md-center p')
        if house_type_description :
            house_type_info ["over_dit_woningentype"] = clean_text("".join(house_type_description.css('::text').getall()))   


        house_type_col = house_type_data.css('th::text').getall()
        house_type_val = house_type_data.css('td').getall()



        house_type_val = [clean_text(i) for i in house_type_val]
        #remove empty spaces in the outcome of 'house_types.css('td::text').getall()'

        house_type_val = [i for i in house_type_val if i !='']

        house_type_col = [normalize_key(i) for i in house_type_col]

        
         #pair the house_type_col and house_type_val up in order
        
        column_counts = {}  # To keep track of the counts for each column nameimport

        # Loop through the columns and values and add them to the dictionary
        for column_name, value in zip(house_type_col, house_type_val):
            # Check if the key already exists in the dictionary
            if column_name in house_type_info:
                # Increment the count for this column
                if column_name in column_counts:
                    column_counts[column_name] += 1
                else:
                    column_counts[column_name] = 1
                # Create a new key with a suffix
                new_key = f"{column_name}_{column_counts[column_name]}"
                house_type_info[new_key] = value
                self.all_possible_house_type_keys.add(new_key)
            else:
                # If the key doesn't exist, add it to the dictionary with the value
                house_type_info[column_name] = value
                self.all_possible_house_type_keys.add(column_name)




        house_type_info['houses'] = house_info_list

        house_type_info_list.append(house_type_info)
            

        # Extract information from the individuale unit   

        houses = response.css('table#tblBouwnummers tbody tr')

        #the number of expected houses found on  the page
        expected_houses = len(houses)
        

        if houses:

                        for house in houses:


                            house_info ={
                                 
                                'house_source':response.url,
                                'bouwnummer':clean_text(house.css('td')[0].get()) if len(house.css('td')) > 0 else None,
                                'prijs':clean_text(house.css('td')[1].get()) if len(house.css('td')) > 1 else None,
                                'kaveloppervlak':'' if clean_text(house.css('td')[2].get()) == "" else clean_text(house.css('td')[2].get()) if len(house.css('td')) > 2 else None,
                                'woonoppervlak':clean_text(house.css('td')[3].get()) if len(house.css('td')) > 3 else None,
                                'slaapkamer':clean_text(house.css('td')[4].get()) if len(house.css('td')) > 4 else None,
                                'inhoud':clean_text(house.css('td')[5].get()) if len(house.css('td')) > 5 else None,
                                'status':clean_text(house.css('td')[6].get()) if len(house.css('td')) > 6 else None,  
                                        
                            }
                
                            house_info_list.append(house_info)
                        
                        if len(house_type_info_list) == expected_house_types:

                            if len(house_info_list) == expected_houses:                    


                                house_type_info_list.append(house_type_info)
                    

        
                                project_info['house_types'] = house_type_info_list

                                yield project_info

        elif len(house_type_info_list) == expected_house_types:
            
            yield project_info

    def get_city_from_url(self, url):
        # Use regex to extract the city name
        match = re.search(r'www\.nieuwbouw(?:-in)?-(\w+)', url)
        if match:
            return match.group(1)
        else:
            return None  # or any default value you prefer

    

    def closed(self, spider):
        # This method is called when the spider is finished

        # Print all possible project keys
        dispatcher.send(signal=signals.spider_closed, sender=spider, all_possible_project_keys=self.all_possible_project_keys,all_possible_house_type_keys=self.all_possible_house_type_keys)




# useful for handling different item types with a single interface
from itemadapter import ItemAdapter
from datetime import datetime
from scrapy.signalmanager import dispatcher
from scrapy import signals

class MergedDataPipeline:
    def __init__(self):
        self.raw_data = []
        self.cleaned_data = []
    def process_item(self, item, spider):
        adapter = ItemAdapter(item)

        if adapter.get('source'):
            #Process raw data
            self.raw_data.append(adapter.asdict())

            #Process cleaned data
            cleaned_item = self.clean_item(item)
            self.cleaned_data.append(cleaned_item)
        return item
    def close_spider(self,spider):

        #Process data
        spider.log("pipeline Closing spider and sending custom_closed_signal...")
        dispatcher.send(signal=signals.spider_closed, 
                        sender=spider, 
                        cleaned_data = self.cleaned_data, 
                        raw_data = self.raw_data)


    def clean_item(self, item):

        if item.get("available_from") =="As soon as possible":
            item["available_from"] = item.get("creation_date")

        # Transform "Available from" field to Unix timestamp
        item['available_from'] = self.convert_date_to_unix_timestamp(item.get('available_from'))

        # Transform "Creation Date" field to Unix timestamp
        item['creation_date'] = self.convert_date_to_unix_timestamp(item.get('creation_date'))

        # Transform boolean values
        bool_keys = ['furnished', 'shareable', 'pets_allowed', 'elevator', 'senior_friendly', 'students_only', 'balcony', 'parking', 'dishwasher', 'washing_machine', 'electric_charging_station', 'dryer']
        for key in bool_keys:
            item[key] = self.to_boolean(item.get(key))

        # Convert numeric fields to integers
        int_keys = ['size', 'rooms','listing-id']
        for key in int_keys:
            item[key] = self.to_integer(item.get(key))

        # Convert prices to float
        float_keys =['monthly_net_rent', 'utilities', 'deposit', 'prepaid_rent', 'move-in_price','housing_deposit']
        for key in float_keys:
            item[key] = self.to_float(item.get(key))

        #Convert "Floor" values to integers
        item['floor'] = self.map_floor_to_integer(item['floor'])

        # Convert "-" to None
        for key,value in item.items():
            if value =='-':
                item[key] = None
        
        self.cleaned_data.append(item)
        return item
    
    def convert_date_to_unix_timestamp(self, date_str):
        try:
            # Specify the format for "01 February 2024"
            date_formats = ["%d %B %Y", "%d %b %Y","%d/%m/%Y"]  # You can add more formats as needed
            for date_format in date_formats:
                try:
                    date_object = datetime.strptime(date_str, date_format)
                    unix_timestamp = int(date_object.timestamp())
                    return unix_timestamp
                except ValueError:
                    pass
            return None  # Return None if no valid format is found
        except Exception as e:
            return None
        
    def to_boolean(self, value):
        if value == "Yes":
            return True
        elif value == "No":
            return False
        else:
            return None

    def to_integer(self, value):
        try:
            return int(value.replace(' m²', '').replace(',', '').replace('.', ''))
        except ValueError:
            return None
        
    def to_float(self, value):
        if value is None:
            return
        try:
            return float(value.replace('.','').replace(',','.').replace(' kr', ''))
        except ValueError:
            return None

    def map_floor_to_integer(self, floor):
        floor = floor.strip().lower()
        if floor == "ground floor":
            return 0
        elif floor[:-2].isdigit():
            return int(floor[:-2])
        else:
            return None
    


# Define here the models for your scraped items
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/items.html

import scrapy


class FinlandrentalscraperItem(scrapy.Item):
    # define the fields for your item here like:
    # name = scrapy.Field()
    pass



import scrapy
from scrapy.exceptions import CloseSpider
import re
import time

def clean_text(text):

    if text is None:

        return '-'

    if text is not None:
                
        # Remove leading and trailing whitespaces, including newline characters
        cleaned_value = text.strip()

        # Replace consecutive newline characters with a single space
        cleaned_value = re.sub(r'\n+', ' ', cleaned_value)

        # Replace other unwanted characters
        cleaned_value = cleaned_value.replace("\u2022", "").replace("\u2013", "").replace("\u2028", "").replace("\u0000", "").replace("\u00A0", "").strip()

        # Use strip() to remove leading and trailing whitespace and newline characters
        return cleaned_value

def normalize_key (text):

    if text is None:
        return ''
    
    if text is not None:
        
        normalized_key = text.replace('(p.m.)','').replace('(s)','').strip()

        # Replace spaces with underscores
        normalized_key = normalized_key.replace(' ', '_').replace('.','').strip()
    
        # Remove special characters
        normalized_key = ''.join(char for char in normalized_key if char.isalnum() or char in ['_', '-'])

        # Decapitalize the key
        normalized_key = normalized_key.lower()

        return normalized_key



class boligportalspiderSpider(scrapy.Spider):
    name = "boligportalspider"
    start_urls = ["https://www.boligportal.dk/en/rental-properties/?offset=0"]
    base_url = "https://www.boligportal.dk/en/rental-properties/?offset={}"

    all_possible_property_keys = set()

    max_retries = 3  # Set the maximum number of retries

    retry_counter = 0


    def parse(self, response):

        #initialize a set for all the possible keys in project_info to keep the amount of the keys for each dict project_info the same

        #Find all the properties
        properties = response.css('div.css-vpsg7t div.temporaryFlexColumnClassName.css-mom5ju div.css-1hh4zdq div.css-1gtq2pq div.temporaryFlexColumnClassName.css-1a08xid')

        last_button = response.css('button.temporaryButtonClassname.css-12fwxlp')[-1]

        is_disabled = last_button.css('::attr(disabled)').get() is not None

        # Extract the current page number from the URL
        current_page = int(response.url.split("offset=")[1])

        # Calculate the next page number with an increment of 18
        next_page_number = current_page + 18

        # Construct the next page URL
        next_page_url = self.base_url.format(next_page_number)

        if is_disabled:

            raise CloseSpider("Reach the last page")


        if not properties:

            if self.retry_counter >= self.max_retries:

                self.logger.warning("Max retries reached, next page...")
                
                self.retry_counter = 0

                yield scrapy.Request(url=next_page_url, callback=self.parse)

            else:

                self.retry_counter += 1

                self.logger.warning("No properties found, waiting and retrying on the same page... (Retry {}/{}))".format(self.retry_counter, self.max_retries))

                time.sleep(0.5)

                yield scrapy.Request(url=response.url, callback=self.parse)

        if properties:

            self.retry_counter = 0    
    
        for property in properties :

            property_info= {}

            #Extract project links (for spiders to crawl into)

            property_link = property.css('div.css-1e7fg19 a ::attr(href)').get()

                        
            if property_link:
                
                yield scrapy.Request(url='https://www.boligportal.dk' + property_link, callback=self.parse_property_page,cb_kwargs={'property_info':property_info})

        
        else:

            # Make a request to the next page
            yield scrapy.Request(next_page_url, callback=self.parse)

    
    def parse_property_page(self,response, property_info= None, retry_attempt=0):


        property_info['source'] = response.url

        property_info['description'] = clean_text(response.css('div.css-1oj64sa div.css-1f7mpex::text').get())

        property_info['address'] = response.css('div.css-v49nss::text').getall()[1]
        
        #parse keys 
        about_property_columns = response.css('div.css-11rix5h section.css-y8cidf:nth-child(1) div.temporaryFlexColumnClassName.css-etn5cp span.css-arxwps::text').getall()[:16]
        #normalize keys
        about_property_columns = [normalize_key (i) for i in about_property_columns]

        property_values = response.css('div.css-11rix5h section.css-y8cidf:nth-child(1) div.temporaryFlexColumnClassName.css-etn5cp span.css-1h46kg2::text').getall()[:16]


        count = 0
        for column in about_property_columns:
            property_info[column] = property_values[count]
            count += 1

        #parse energy_label info
        property_info['energy_label'] = ()
        if response.css('div[tabindex="-1"][data-state="closed"] img.css-rdsunt::attr(src)').get():

            property_info['energy_label'] = response.css('div[tabindex="-1"][data-state="closed"] img.css-rdsunt::attr(src)').get().split("/")[-1].split("_")[0]

        else:
            property_info['energy_label'] = '-'

        #parse the keys
        about_rental_columns = response.css('div.css-11rix5h section.css-y8cidf:nth-child(2) div.temporaryFlexColumnClassName.css-etn5cp span.css-arxwps::text').getall()

        #normalize keys
        about_rental_columns = [normalize_key (i) for i in about_rental_columns]

        rental_values = response.css('div.css-11rix5h section.css-y8cidf:nth-child(2) div.temporaryFlexColumnClassName.css-etn5cp span.css-1h46kg2::text').getall()

        count = 0
        for column in about_rental_columns:
            property_info[column] = rental_values[count]
            count += 1



        if not property_info and retry_attempt < 1 :
            # Raise an exception to trigger the errback function
            raise Exception("Conditions not met")

        yield property_info

def retry_request(self, failure):
    if failure.check(Exception) and "Retry request" in str(failure.value):
        retry_attempt = failure.request.cb_kwargs.get('retry_attempt', 0)
    # Handle retries here
        if failure.check(Exception):
            # Reload the same page and retry the request
            url = failure.request.url
            self.logger.warning("Retrying the request for URL: %s (Retry %d)", url,retry_attempt + 1)
            yield scrapy.Request(
                url=url,
                callback=self.parse_property_page,
                cb_kwargs={'property_info': failure.request.cb_kwargs['property_info'],'retry_attempt': retry_attempt + 1},
                errback=self.retry_request  # Continue retrying
            )





# Scrapy settings for nieuwbouwscraper project
#
# For simplicity, this file contains only settings considered important or
# commonly used. You can find more settings consulting the documentation:
#
#     https://docs.scrapy.org/en/latest/topics/settings.html
#     https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
#     https://docs.scrapy.org/en/latest/topics/spider-middleware.html

settings = {
    "BOT_NAME": "nieuwbouwscraper",
    "SPIDER_MODULES": ["ClappScrapers.nieuwbouw.spider"],
    "NEWSPIDER_MODULE": "ClappScrapers.nieuwbouw.spider",
    "ROBOTSTXT_OBEY": True,
    "ITEM_PIPELINES": {
        "ClappScrapers.nieuwbouw.spider.pipelines.MergedDataPipeline": 100,
    },
    "DOWNLOADER_MIDDLEWARES":{
        "scrapy.downloadermiddlewares.useragent.UserAgentMiddleware":None,
        "scrapy_user_agents.middlewares.RandomUserAgentMiddleware":700,
    },
    "REQUEST_FINGERPRINTER_IMPLEMENTATION": "2.7",
    "TWISTED_REACTOR": "twisted.internet.asyncioreactor.AsyncioSelectorReactor",
    "FEED_EXPORT_ENCODING": "utf-8",
    "CUSTOM_LOG_EXTENSION":True,
    "EXTENSIONS":{
        'scrapy.extensions.telnet.TelnetConsole': None,
        'ClappScrapers.nieuwbouw.spider.extension.CustomLogExtension': 1,}
}




settings = {
  "BOT_NAME": "finlandrentalscraper",
  "SPIDER_MODULES": ["ClappScrapers.fin_rental.spider"],
  "NEWSPIDER_MODULE": "ClappScrapers.fin_rental.spider",
  "ROBOTSTXT_OBEY": True,
  "ITEM_PIPELINES": {
      "ClappScrapers.fin_rental.spider.pipelines.MergedDataPipeline": 100,
  },
  "DOWNLOADER_MIDDLEWARES": {
      "scrapy.downloadermiddlewares.useragent.UserAgentMiddleware": None,
      "scrapy_user_agents.middlewares.RandomUserAgentMiddleware": 700,
  },
  "AUTOTHROTTLE_ENABLED": True,
  "AUTOTHROTTLE_START_DELAY": 0.5,
  "AUTOTHROTTLE_TARGET_CONCURRENCY": 1.0,
  "REQUEST_FINGERPRINTER_IMPLEMENTATION": "2.7",
  "TWISTED_REACTOR": "twisted.internet.asyncioreactor.AsyncioSelectorReactor",
  "FEED_EXPORT_ENCODING": "utf-8",
  "CUSTOM_LOG_EXTENSION":True,
  "EXTENSIONS":{
    'scrapy.extensions.telnet.TelnetConsole': None,
    'ClappScrapers.fin_rental.spider.extension.CustomLogExtension': 1,}
}